import polars as pl
import shutil
import typer
import os
from pathlib import Path
from typing import List, Tuple, Dict, Union, Optional, Any, Iterator, TypeVar

from iotree.core.render.funcs import basic_pbar, rich_func_chainer


from pisql.core.isql import iSQL, check_sql, iSqlDF
from pisql.core.isql_config import init_isql, iSqlCFG
from pisql.utils.render import console, errors

from pisql.utils.paths import (
    base_dir, tests_dir, package_dir, assets_dir, data_dir, pisql_home
)

isql = init_isql()

run = typer.Typer(
    name="run",
    help=f"""Activates query mode: execute queries  against a Sybase ASE database . Is able to process many .sql files  and even directories ",  
    [yellow][bold]Aliases:[/] q , :: [/]  

    [green][bold]Example:[/]  

    >>> pisql q ++ name.sql cool.sql[/] [dim]# Execute multiple files[/]
    """,
    rich_markup_mode="rich",
    rich_help_panel="rich",
)

def splitSqlScript(sqlScriptFile: Union[str, Path, os.PathLike]) -> List[Path]:
    """Split a SQL script into a list of queries"""
    sqlScriptFile = Path(sqlScriptFile)
    raw = open(sqlScriptFile, "r").read()
    queries = raw.replace("go", "GO").split("GO")

    if not isinstance(queries, list):
        return [sqlScriptFile]
    elif len(queries) <= 2:
        if not all( check_sql(query) for query in queries):
            return [sqlScriptFile]

    queries = [query.strip() for query in queries if check_sql(query.strip()) ]
    filenames = [sqlScriptFile.stem+"-subquery-"+str(idx)+".sql" for idx in range(len(queries))]
    
    tmpdir = pisql_home.joinpath('tmp')
    tmpdir.mkdir(exist_ok=True)

    pathnames = [tmpdir.joinpath(filename) for filename in filenames]
    for pathname, query in zip(pathnames, queries):
        with open(pathname, "w+") as f:
            f.write(query)
    return pathnames

def cleanSplitScriptDir(pathnames: List[Path]) -> None:
    """Clean up the temporary directory of split SQL scripts"""
    for pathname in pathnames:
        os.remove(pathname)
    tmpdir = pisql_home.joinpath('tmp')
    shutil.rmtree(tmpdir)


@run.command(name="//", help="Execute a query  against a Sybase ASE database . Accepts a dir  of .sql files  or a list of files")
def runall(
    args : List[str],
    outdir: Path = typer.Option(None, "--outdir", "-o", help="Output directory for results"),
    ) -> None:
    """Execute a query  against a Sybase ASE database . Accepts a dir  of .sql files  or a list of files"""
    files = []
    tmpfiles = []

    outdir = pisql_home / "results" if outdir is None else outdir
    outdir.mkdir(exist_ok=True)

    for arg in args:
        arg = Path(os.getcwd() + arg[1:] if arg[0] == "." else arg)
        if arg.is_dir():
            files += [file for file in arg.iterdir() if file.suffix == ".sql"]
        elif arg.is_file():
            splitRes = splitSqlScript(arg)
            if splitRes is None:
                files.append(arg)
            else:
                files += splitRes
                if len(splitRes) > 1:
                    tmpfiles += splitRes
        else:
            raise ValueError(f"Invalid file or directory : {arg}")
    
    last_idx = 0
    callback = lambda *args : "subquery" in args[0].name
    for idx, df in rich_func_chainer(isql.run_sql_file, files, console=console, condition_innerloop=callback):
        if idx is None:
            errors.print(f"🔥 Error: {df}")
            write_idx = last_idx+1 if last_idx > 0 else 0
            open(outdir / f"error-{files[write_idx].stem}.log", "w+").write(str(df))
            continue
        else:
            df.write_parquet( outdir / f"results-{files[idx].stem}.parquet", compression="lz4")
            last_idx = idx
    
    cleanSplitScriptDir(tmpfiles)
    console.print(f"✅ Done, Results saved to {pisql_home / 'results'}")


@run.command(name="++", help="Execute a list of queries  against a Sybase ASE database . Accepts a dir  of .sql files  or a list of files")
def runallAlias(
    args : List[str],
    ) -> None:
    """Execute a query  against a Sybase ASE database . Accepts a dir  of .sql files  or a list of files"""
    runall(args=args)


def listTablesServer() -> None:
    """List all tables  in the server """
    querytxt = """
    SELECT t.name as table_name, c.name as column_name
    FROM sysobjects t, syscolumns c
    WHERE t.id = c.id AND t.type IN ('U', 'V')
    """
    df = isql.query(querytxt)
    console.print(df)
    df.write_parquet(pisql_home / "results" / "tables.parquet", compression="lz4")

@run.command(name="tables?", help="List all tables  in the server . [bold yellow]Alias: `?`[/]")
def listTables() -> None:
    """List all tables  in the server """
    listTablesServer()

@run.command(name="?", help="List all tables  in the server ", hidden=True)
def listTables() -> None:
    """List all tables  in the server  [bold yellow]Alias: `tables?`[/]"""
    listTablesServer()



from .core.isql_config import init_isql
from .core.isql import iSQL, iSqlDF, check_sql

from .utils.paths import (
    base_dir, tests_dir, package_dir,
    assets_dir
)

from .utils.render import console, errors

from .utils.local_store import iSqlConfig

from .cli.run import run
from .cli.execute import exe
from .cli.config import config


from rich.console import Console

console = Console(markup=True, color_system="auto", emoji=True, record=True)

errors = Console(
    markup=True, color_system="auto",
    emoji=True, record=True, stderr=True)


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pisql', 'pisql.cli', 'pisql.core', 'pisql.utils']

package_data = \
{'': ['*']}

install_requires = \
['iotree==0.1.21-post1', 'polars>=0.15.13,<0.16.0', 'typer[all]>=0.7.0,<0.8.0']

entry_points = \
{'console_scripts': ['pisql = pisql.main:app']}

setup_kwargs = {
    'name': 'pisql',
    'version': '0.1.31.post10',
    'description': 'pisql is a CLI + lightweight python library to interact with a Sybase ASE database.',
    'long_description': '[![License](https://img.shields.io/badge/Licence-MIT-blue.svg)](/LICENSE) [![PyPI](https://img.shields.io/badge/PyPI-latest-green.svg)](https://pypi.org/project/pisql) [![License](https://img.shields.io/badge/Archives-targz-purple.svg)](https://pypi.org/project/pisql/#files)\n\n# What is pisql ?\n\npisql is a mix of a command line tool and a small python library to interact with a [Sybase ASE][1] database.  \nIt so happens, that I have to work with a Sybase ASE database and I wanted to have a tool to interact with it.  \n\n## But why ?\n\nThe reason I built this is fairly simple although frustrating:\n\n* My company has tight control over what gets installed on the workstations\n* My company hasn\'t bought the SQL Anywhere drivers for ASE on top of ASE itself as of today\n* The only way I have to interact with the database is through a [very old and ugly piece of software called "sqlDbx"][2].\n* I wanted to have scripting and automation capabilities that the interface couldn\'t provide, but that bash/powershell/python could.\n\n## What is it ?\n\npisql is, for the CLI part, a [rich CLI wrapper][3] around the barebones [isql][4] command line tool that comes with the ASE installation every time.\n\nFor the python part, it\'s a small library that allows you to interact with the database through python code.\nYou can turn `.sql` files into dataframes ([pandas](https://pandas.pydata.org/) or [polars](https://pola.rs)), and further manipulate them.\nYou can use every tool you have in python to interact with the data, vizualize it, etc.\n\n## How do I use it ?\n\n### Installation\n\nYou can install pisql through pip:\n\n```bash\n    pip install pisql\n```\n\nAlthough I recommend `pipx`:\n\n```bash\n    pipx install pisql\n```\n\nThis works on Windows, Linux and MacOS.\n\n### Usage\n\n#### CLI\n\nThe CLI is fairly simple to use.\n\nTo execute a single .sql file, you can just use the `exec` command:\n\n```bash\n    pisql exec my_file.sql\n```\n\nwhich is also aliased to `pisql e my_file.sql` and `pisql x my_file.sql`.\n\nTo execute multiple .sql files, you can use what I call the "query mode" or "run mode", using either  \nthe symbols `q`, `run` or `::`. Once this more is activated, you can chain multiple executions together  \nby using the `++` or `//` commands:\n\n```bash\n    pisql q ++ file_one.sql file_two.sql file_three.sql\n```\n\nAn important feature of this mode, is that you can list both files and directories.\n\n\n```bash\n    pisql :: // file.sql some_dir other_file.sql\n```\n\nWhat happens then is that pisql will execute sequentially:\n\n* `file.sql` first\n* then all the `.sql` files in `some_dir` second\n* then `other_file.sql` last\n\nOne nice feature is the presence of rich progress bars like so (give example)\n\n**NB:** I haven\'t had the time to implement further recursion, so if you have a directory in `some_dir`, it will be ignored.\n\n#### Python\n\nWill explain in the next few days.\n\n## What\'s next ?\n\nI\'m currently working on a few things:\n\n* [ ] Implementing a config file subcommand to set the default database, user, etc.\n* [ ] Give more freedom to users to change the storage of the dataframes and config files\n* [ ] Build a semi ORM to interact with the database through python. Comes to mind the SELECT, WHERE, JOIN, etc. clauses.\n* [ ] Have a little templating feature, but nothing too fancy. I don\'t want to reinvent the wheel here.\n* [ ] Use Agg-Grid to have a  web interface to interact with the database. I\'m not sure how to do this yet, but I\'ll figure it out.\n\nHere\'s most of it for now ! I\'ll update this as I go.\n\n[1]: https://www.sap.com/products/technology-platform/sybase-ase.html\n[2]: http://www.sqldbx.com/\n[3]:https://github.com/Textualize/rich\n[4]:https://infocenter.sybase.com/help/index.jsp?topic=/com.sybase.infocenter.dc35456.1570/html/ocspsunx/X33477.htm',
    'author': 'Arno V',
    'author_email': 'arno@veletanlic.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
